module vga10 (clk,r,g,b,hs,vs,vgaclk,       // vga interface
             addr,datain,dataout,wr,rd,ready,      // cpu interface
             sram_addr,sram_ce_n,sram_dq,sram_lb_n,sram_oe_n,sram_ub_n,sram_we_n,
				 debug,hbl,vbl);// sram interface); 
				 // 1440x900 160x50 char with border text mode vga driver for DE2-115

// ver 0.15 alpha --- 2014.09.03
// (c) Piotr Kardasz pik33@we.pb.edu.pl
// This is free software, license: LGPL v.2.0 or newer

output reg [7:0] r,g,b;
output reg hs,vs;
output wire vgaclk; 
input wire clk;  //input clock=2x vgaclk
output wire ready;

output reg[7:0] debug;

reg [7:0] e;
reg f,oldwr,oldwr2;
reg [7:0] brr,brb,brg;
reg [19:0] caddr;
reg [31:0] lcnt;
reg [0:2] lstate,pstate; // 00 vis, 01 fp, 10 sync, 11 bp 100 border
reg [11:0] pcnt;
reg [31:0] fcnt; //frame counter

input wire [20:0] addr;
input wire [31:0] datain;
output reg [31:0] dataout;
input wire wr,rd;
output reg [19:0] sram_addr;
inout wire [15:0] sram_dq;
output wire sram_ce_n;
output reg sram_lb_n;
output wire sram_oe_n;
output reg sram_ub_n;
output reg sram_we_n;
output reg hbl,vbl;
reg [15:0] sramdqout,dataout2,dataout3;
reg[31:0] datain2,datain3;
reg [20:0] addr2,sram_addr2;
reg[3:0] pcnt_mod,pcnt_mod2;
reg[7:0] pcnt_char;
reg[15:0] pixels3, pixels2;
reg[31:0] pixels,bb,cc,ss,mode;
reg wr2,wr3,rd2,rd3,oldrd;
reg[3:0]wstate;


(* ramstyle = "M9K" *) reg [31:0] pallette[256]; // pallette registers
(* ramstyle = "M9K" *) reg [7:0] fontdef[2048];  // character generator
(* ramstyle = "M9K" *) reg [31:0] sprite[2048];  // sprite generator
reg[31:0] regs[4];  // internal regs

// 0 - display state: {vsync,hsync,frame counter}
// 1 - display start address
// 2 - mode
//
// 15 - border color 
// 16..31: pallette

initial
  begin
  $readmemh("st4font.hex",fontdef);
  end

  
assign vgaclk=clk;
assign sram_ce_n=0;
assign sram_oe_n=0;
 

// ------------------------------------------- 
// bidir assignment
// ------------------------------------------- 

assign sram_dq = (sram_we_n==0) ? sramdqout : 16'bz;

//---------------------------------------------  
// debug  
//---------------------------------------------  
  
always@(posedge clk) 
 
begin 

debug[3:0]<=wstate;
debug[7]<=wr;
debug[6]<=wr3;
debug[5]<=rd;
debug[4]<=rd3;
end



//---------------------------------------------
// main loop
//--------------------------------------------- 
 
always@(posedge clk)

  begin
  oldwr<=wr;
  if ((oldwr==0) && (wr==1)) wr2<=1; else wr2<=0;
  wr3<=wr2;
  oldrd<=rd;
  if ((oldrd==0) && (rd==1)) rd2<=1; else rd2<=0;
  rd3<=rd2;
  pcnt_mod2<=pcnt_mod-1;
  
// ------------------------------------------- 
// sram rw state machine
// -------------------------------------------
  
// new state calculation
  case (wstate) 
  
    0: begin
	   addr2<=addr;
		if (((wr2==1) || (wr3==1)) && (rd2==0) && (rd3==0) && (addr[19]==0)) wstate<=1;
		else if (((wr2==1) || (wr3==1)) && (rd2==0) && (rd3==0) && (addr[19]==1)) wstate<=3;
		else if (((rd2==1) || (rd3==1)) && (wr2==0) && (wr3==0) && (addr[19]==0)) wstate<=4;
		else if (((rd2==1) || (rd3==1)) && (wr2==0) && (wr3==0) && (addr[19]==1)) wstate<=5;
		else wstate<=0;
     end
	  
	1: begin  
	 
		if ((pcnt_mod2==0) || (pcnt_mod2==1)) wstate<=6; else wstate<=1; 
		datain2=datain;
		sramdqout<=datain2[31:16];
		sram_addr2<={addr,1'b0};
      end 

   6: begin		
	
		if (pcnt_mod2==4) wstate<=7; else wstate<=6; 
  	   sram_addr<=sram_addr2;
	   sram_we_n<=0;    
      sram_lb_n<=0;
      sram_ub_n<=0;  
      end

	7: begin

		if (pcnt_mod2==5) wstate<=8; else wstate<=7;
      sram_lb_n<=1;
      sram_ub_n<=1;	 
   	sram_addr<=sram_addr2+1;
  	   sramdqout<=datain2[15:0];
	   end
  
   8: begin
  
      if (pcnt_mod2==7) wstate<=9; else wstate<=8;
      sram_lb_n<=0;
      sram_ub_n<=0;	 
  	   end	  
  
   9: begin
  
		if (pcnt_mod2==8) wstate<=0; else wstate<=9;
   	sram_we_n<=1;    
      sram_lb_n<=1;
      sram_ub_n<=1;  
      end

	4:	begin
	 
		if (pcnt_mod2==0) wstate<=10; else wstate<=4;
		sram_addr2<={addr,1'b0};
      end		
		
	10:begin	
	   case (pcnt_mod2)
      1,2,3:
		    begin
	       sram_addr<=sram_addr2;
	       sram_we_n<=1;    
          sram_lb_n<=0;
          sram_ub_n<=0;  
 	       end
		4:
         begin
   	   sram_addr<=sram_addr2+1;
	      dataout3<=sram_dq;
  	      sram_we_n<=1;    
         sram_lb_n<=0;
         sram_ub_n<=0;  
	      end 	 

	   5,6,7:begin
   	   sram_addr<=sram_addr2+1;
  	      sram_we_n<=1;    
         sram_lb_n<=0;
         sram_ub_n<=0;  
	      end		
		
   	8:
         begin
	      dataout2<=sram_dq;
	      end 	 			
		9:
		   begin
		   dataout<={dataout3,dataout2};	 
	      wstate<=0;
			end
	   default:
         wstate<=0;
	      		
			
		endcase	
      end
		
   3: begin  
      wstate<=0;
      if (addr[15:12]==4'b0000) pallette[addr2[7:0]]<=datain; 
		else if (addr[15:12]==4'b0001)fontdef[addr2[10:0]]<=datain[7:0];
		else if (addr[15:12]==4'b0010)sprite[addr2[10:0]]<=datain;
		else regs [addr2[10:0]]<=datain;
   	end
		
	5: begin	

  	   wstate<=0;
	   if (addr[15:12]==4'b0000) dataout<=pallette[addr2[7:0]]; 
		else if (addr[15:12]==4'b0001) dataout<={24'b0,fontdef[addr2[10:0]]};
		else if (addr[15:12]==4'b0010) dataout<=sprite[addr2[10:0]];
		else dataout<=regs [addr[10:0]];
	   end
	 
   default: wstate<=0;
	
	endcase
	
//----- end of sram interface state machine

//----- vga 
	 
  if (pcnt_mod==9)	 	 
    begin
	 sram_addr<=caddr;
	 sram_we_n<=1;    
    sram_lb_n<=0;
    sram_ub_n<=0;  
	 end 
	 
  if (pcnt_mod==12)	 	 
    begin
	 sram_addr<=caddr+1;
	 pixels2<=sram_dq;
	 sram_we_n<=1;    
    sram_lb_n<=0;
    sram_ub_n<=0;  
	 end 

  if (pcnt_mod==15)	 	 
    begin
	 pixels3<=sram_dq;

	 end 	 
	 
  if (pcnt_mod==0) 
    begin
	 pixels={pixels2,pixels3};
    end
	



  case (pcnt)
 
    33:ss<=regs[1];
    34:bb<=regs[3];
    35:mode<=regs[2];
    36:regs[0]<=lcnt;

  endcase 

  pcnt<=pcnt+1;
  pcnt_mod<=(pcnt+1)&15;
  pcnt_char<=(pcnt-144)>>4;
    // line and pixel counters & state determination
  if (pcnt==79+16) begin pstate<=4; hbl<=0; end  // left border
  if (pcnt==143+16) pstate<=0; // display
	 if (pcnt==1935+16) begin pstate<=4; end;// right border
	 if (pcnt==1999+16) begin pstate<=1; hbl<=1; end // front porch
	 if (pcnt==2047+16) begin pstate<=2; hs<=0; end// hsync
	 if (pcnt==2079+16)           // back porch 
	   begin 
		pstate<=3; 
      hs<=1;
		pcnt<=16;
		lcnt<=lcnt+1; 

		if (lcnt==25) begin vbl<=0; lstate<=4; end   // top border
		if (lcnt==65) lstate<=0;   // display
		if (lcnt==1185) lstate<=4;  // bottom border
		if (lcnt==1225) begin lstate<=1; vbl<=1;  end  // front porch v
		if (lcnt==1228) begin lstate<=2; vs<=1; end // vsync 
		if (lcnt==1234) begin lstate<=3; vs<=0; lcnt<=0; fcnt<=fcnt+1; end // back porch v
		end
 
    caddr=((lcnt-66)>>mode)*224+(pcnt_char<<1)+(ss<<1);
	 

	 if (pstate==1)
	   begin
		brr=bb[7:0];
		brg=bb[15:8];
		brb=bb[23:16];
      end 
    
    if (((lstate==4)&&((pstate&3)==0)) || ((pstate==4)&&(lstate==0))) // borders
      begin 
      r<=brr; g<=brg; b<=brb;	
      end		 
	 
    else if (lstate==0 && pstate==0) // active display
      begin


		if (mode==0)
        case (pcnt_mod)
          0:e=pixels[31:30];
	  	    1:e=pixels[29:28];
		    2:e=pixels[27:26];
		    3:e=pixels[25:24];
		    4:e=pixels[23:22];
		    5:e=pixels[21:20];
		    6:e=pixels[19:18];
		    7:e=pixels[17:16];		            
			 8:e=pixels[15:14];
	  	    9:e=pixels[13:12];
		    10:e=pixels[11:10];
		    11:e=pixels[9:8];
		    12:e=pixels[7:6];
		    13:e=pixels[5:4];
		    14:e=pixels[3:2];
		    15:e=pixels[1:0];		  
		  endcase
		else if (mode==1)  
		   case (pcnt_mod)
  		   0,1:e=pixels[31:28];
		   2,3:e=pixels[27:24];
		   4,5:e=pixels[23:20];
		   6,7:e=pixels[19:16];  		   
			8,9:e=pixels[15:12];
		   10,11:e=pixels[11:8];
		   12,13:e=pixels[7:4];
		   14,15:e=pixels[3:0];
       endcase
	
   	else   
		   case (pcnt_mod)
  		   0,1,2,3:e=pixels[31:24];
		   4,5,6,7:e=pixels[23:16];
		   8,9,10,11:e=pixels[15:8];
		   12,13,14,15:e=pixels[7:0];  		   
        endcase


		cc<=pallette[e]; 
      r=cc[7:0]; g=cc[15:8]; b=cc[23:16]; 
		end
	 
	 else begin r<=0; g<=0; b<=0; end		

  end
 
endmodule 